/**
 * Copyright 2006-8 Simon Andrews
 *
 *    This file is part of ChIPMonk.
 *
 *    ChIPMonk is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    ChIPMonk is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with ChIPMonk; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package uk.ac.bbsrc.babraham.ChipMonk.DataParsers;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.Enumeration;
import java.util.Vector;

import uk.ac.bbsrc.babraham.ChipMonk.ChipMonkApplication;
import uk.ac.bbsrc.babraham.ChipMonk.ChipMonkException;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.Chromosome;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.DataCollection;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.DataSet;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.Probe;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.ProbeSet;

public class NimblegenParser implements Runnable {

	private ChipMonkApplication application;
	private Vector listeners = new Vector();
	private ProbeSet probeSet;
	private DataCollection dataCollection;
	private BufferedReader br;
	private NimblegenSamplePair [] sampleIndices;
	private DataSet [] dataSets;
	
	
	public NimblegenParser (ChipMonkApplication application) {
		this.application = application;
		dataCollection = application.dataCollection();
		probeSet = dataCollection.probeSet();
	}
	
	public ChipMonkApplication application () {
		return application;
	}
	
	public void parseFile (File file) {
	
		// We need to collect information about which sections are which.  We can get this
		// from the first file.  We can assume that all pair files 

		try {
			br = new BufferedReader(new FileReader(file));
			
			String line = br.readLine();
			
			if (line == null) {
				throw new ChipMonkException("Data file "+file.getName()+" was empty");
			}
			
			String [] sections = line.split("\\t");
//			for (int i=0;i<sections.length;i++) {
//				System.out.println(i+" : "+sections[i]);
//			}
			
			// The heading sections in a Niblegen format file are:
			// 0 : GENE_EXPR_OPTION
			// 1 : SEQ_ID
			// 2 : PROBE_ID
			// 3 : POSITION
			
			if (sections.length < 6) {
				throw new ChipMonkException("Not enough header fields ("+sections.length+") to be a Numblegen all_pairs file");
			}
			
			// Just to confuse things a bit - some Nimblegen files have an extra
			// set of headers before the data.  So the data either starts at index
			// 4 or index 7.  The extra fields are:
			// 4 : X
			// 5 : Y
			// 6 : MATCH_INDEX
			
			int dataStartIndex = 4;
			if (sections[4].equals("X") && sections[5].equals("Y")) {
				dataStartIndex = 7;
			}
//			System.out.println("Data start index is "+dataStartIndex+" length of line is "+sections.length);

			
			// This next check was causing problems since Nimblegen keep adding new
			// fields to their all_pairs file.  In the end it's going to be
			// easier to show all possible columns (minus the obvious headers) and
			// let the user pick out the real ones. If they get it wrong they'll just
			// end up with a parser error anyway.
			
//			if ((sections.length-dataStartIndex)%2 != 0) {
//				throw new ChipMonkException("Not an even number of measures in the file, can't create pairs");
//			}
			
			NimblegenPairSelector nps = new NimblegenPairSelector(this,sections,dataStartIndex);
			sampleIndices = nps.getSamples();
			if (sampleIndices == null) {
				// The user didn't define any samples or cancelled
				
				Enumeration e = listeners.elements();
				while (e.hasMoreElements()) {
					((DataParserListener)e.nextElement()).dataProcessingFinished(true);
				}
				return;
			}
			dataSets = new DataSet [sampleIndices.length];
			for (int i=0;i<sampleIndices.length;i++) {
				dataSets[i] = new DataSet(sampleIndices[i].name);
			}
		}
		catch (Exception ex) {
			Enumeration e = listeners.elements();
			while (e.hasMoreElements()) {
				((DataParserListener)e.nextElement()).dataExceptionReceived(ex);
			}
			return;
		}
		
		
		Thread t = new Thread(this);
		t.start();
	}
	
	public void addDataParserListener (DataParserListener l) {
		listeners.add(l);
	}
	
	public void run() {
		// Parse the files...
		int count = 0;
		try {
			String line;
			while ((line = br.readLine()) != null) {
				count++;
				
				if ((count % 1000)==0) {
					Enumeration e = listeners.elements();
					while (e.hasMoreElements()) {
						((DataParserListener)e.nextElement()).dataUpdateProgress(0,1,"Loaded data for "+count+" probes");
						try {
							Thread.sleep(5);
						}
						catch (InterruptedException ie){}
					}
				}
				if (line.length() == 0) continue;
				
				String [] sections = line.split("\\t");
				
				// Control probes start with RANDOM and can be ignored.
				if (sections[2].startsWith("RANDOM"))
					continue;

				Probe p = application.dataCollection().probeSet().getProbeWithName(sections[2]);
				
				if (p == null) {
					// Probe names look like CHR1P2345, where 1 is the chromosome
					// name and 2345 is the position.  Some files also use FS instead of P (no idea why?)
					String [] probeSections = sections[2].split("(P|FS)");
					if (probeSections.length != 2) {
						throw new ChipMonkException("Couldn't get chr and position from name "+sections[2]);
					}
	
					// The filters here are for later Nimblegen formats where they padded the
					// chromosome name and position with 0000.  Position is fine (leading pads)
					// but they padded the name at the end and the beginning (!) so you get things
					// like CHR0100P002345.
					
					
					String chrName = probeSections[0];
					// We want only two chars, but if it's an older format we only need one for
					// single char chromosome names.
					if (chrName.length()> 4) {
						chrName = chrName.substring(3,5);						
					}
					else {
						chrName = chrName.substring(3,chrName.length());						
					}
					while (chrName.charAt(0) == '0') {
						chrName = chrName.substring(1);
					}

					
					// They also treat letters (X/Y) differently to numbers.  CHR1 is 01 but
					// X is X0. We therefore need to strip trailing 0s from non-digit chromosomes.

					if (! Character.isDigit(chrName.charAt(0))) {
						while (chrName.charAt(chrName.length()-1) == '0') {
							chrName = chrName.substring(0,chrName.length()-1);
						}
					}
					
					try {
						Chromosome c = application.genome().getChromosome(chrName);
						int position = Integer.parseInt(probeSections[1]);
						if (c == null) {
							Enumeration e = listeners.elements();
							while (e.hasMoreElements()) {
								((DataParserListener)e.nextElement()).dataWarningReceived(new ChipMonkException("Probe "+sections[2]+" could not be mapped to chromosome '"+chrName+"'"));
							}
						}
						else if (position > c.length()) {
							Enumeration e = listeners.elements();
							while (e.hasMoreElements()) {
								((DataParserListener)e.nextElement()).dataWarningReceived(new ChipMonkException("Probe "+sections[2]+" has a position "+position+" beyond the end of chr"+c.name()+" ("+c.length()+")"));
							}
							c = null;
							position = 0;
						}

						p = new Probe(sections[2],c,position);
					}
					catch (ChipMonkException ex) {
						ex.printStackTrace();
						continue;
					}
					probeSet.addProbe(p);
				}
				
				for (int i=0;i<dataSets.length;i++) {
					dataSets[i].addData(p,Double.parseDouble(sections[sampleIndices[i].signal]),Double.parseDouble(sections[sampleIndices[i].control]));
				}
			}
			
			br.close();
		}
		catch (Exception ex) {
			Enumeration e = listeners.elements();
			while (e.hasMoreElements()) {
				((DataParserListener)e.nextElement()).dataExceptionReceived(ex);
			}
			return;
		}
		
		for (int i=0;i<dataSets.length;i++) {
			dataCollection.addDataSet(dataSets[i]);
		}
		
		Enumeration e = listeners.elements();
		while (e.hasMoreElements()) {
			((DataParserListener)e.nextElement()).dataProcessingFinished(true);
		}
	}

}
