/**
 * Copyright 2006-8 Simon Andrews
 *
 *    This file is part of ChIPMonk.
 *
 *    ChIPMonk is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    ChIPMonk is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with ChIPMonk; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package uk.ac.bbsrc.babraham.ChipMonk.DataTypes;

import java.io.File;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import uk.ac.bbsrc.babraham.ChipMonk.ChipMonkException;

public class Genome {

	private String species;
	private String assembly;
	
	private Hashtable chromosomes = new Hashtable();
	
	private Vector unLoadedFeatureTypes = new Vector();
	
	
	public Genome (File baseLocation) throws ChipMonkException {
		if (!baseLocation.exists()) {
			throw new ChipMonkException("Couldn't find the "+baseLocation.getName()+" genome at "+baseLocation.getAbsolutePath()+".  You may need to add it to your Genome Base location");
		}
		if (!baseLocation.isDirectory()) {
			throw new ChipMonkException("Base location for new genome must be a directory: "+baseLocation.getAbsolutePath());
		}
		
		String [] sections = baseLocation.getAbsolutePath().split("[\\\\\\/]");
		if (sections.length < 2) {
			throw new ChipMonkException("Couldn't identify the species and assembly from "+baseLocation.getAbsolutePath());
		}
		assembly = sections[sections.length-1];
		species = sections[sections.length-2];
	}
	
	public void addChromosome (String name) throws ChipMonkException {
		if (chromosomes.contains(name)) {
			throw new ChipMonkException("Duplicate chromosome name "+name+" is not allowed");
		}
		chromosomes.put(name,new Chromosome(name));
	}
	
	public int getChromosomeCount (){
		return chromosomes.size();
	}

	public boolean hasChromosome (String name) {
		if (chromosomes.containsKey(name)) {
			return true;
		}
		else {
			return false;
		}
	}
	
	public Chromosome [] getAllChromosomes () {
		
		Chromosome [] c = new Chromosome [chromosomes.size()];
		
		Enumeration e = chromosomes.keys();
		
		int count = 0;
		while (e.hasMoreElements()) {
			c[count] = (Chromosome)chromosomes.get(e.nextElement());
			++count;
		}
		
		Arrays.sort(c);
		
		return c;
	}
	
	public Chromosome getChromosome (String name) {
		try {
			return ((Chromosome)chromosomes.get(name));			
		}
		catch (NullPointerException e) {
			// We get this if our key is null
			return null;
		}
	}
	
	public String [] listUnloadedFeatureTypes () {
		return (String [])unLoadedFeatureTypes.toArray(new String[0]);		
	}
	
	public void addUnloadedFeatureType (String name) {
		if (! unLoadedFeatureTypes.contains(name)) {
			unLoadedFeatureTypes.add(name);
		}
	}
	
	public String [] listAvailableFeatureTypes () {
		Vector t = new Vector();
		Enumeration e = chromosomes.keys();
		while (e.hasMoreElements()) {
			String [] types = ((Chromosome)chromosomes.get(e.nextElement())).listAvailableFeatureTypes();
			for (int i=0;i<types.length;i++) {
				if (! t.contains(types[i])) {
					t.add(types[i]);
				}
			}
		}
		return (String [])t.toArray(new String[0]);
	}
	
	public String species () {
		return species;
	}
	
	public String assembly () {
		return assembly;
	}

	public int getLongestChromosomeLength () {
		int l = 0;
		Chromosome [] c = getAllChromosomes();
		for (int i=0;i<c.length;i++) {
			if (c[i].length() > l) {
				l=c[i].length();
			}
		}
		return l; 
	}
}
