/**
 * Copyright 2006-8 Simon Andrews
 *
 *    This file is part of ChIPMonk.
 *
 *    ChIPMonk is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    ChIPMonk is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with ChIPMonk; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package uk.ac.bbsrc.babraham.ChipMonk.Displays.ChromosomeViewer;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.Enumeration;
import java.util.Vector;

import javax.swing.JPanel;

import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.DataCollection;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.DataStore;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.Probe;
import uk.ac.bbsrc.babraham.ChipMonk.Gradients.ColourGradient;

public class ChromosomeDataTrack extends JPanel {
	
	private ChromosomeViewer viewer;
	private DataStore data;
	
	private Probe [] probes;
	private int x;
	private int width;
	private int fullWidth;
	private int height;
	private Vector drawnProbes = new Vector();
	private Probe activeProbe = null;
	private double maxLogRatio;
	private double minLogRatio;
		
	// These are used to optimise the drawing routines
	private int lastXStart = 0;
	
	public ChromosomeDataTrack (ChromosomeViewer viewer, DataCollection collection, DataStore data) {
		this.viewer = viewer;
		this.data = data;
		probes = collection.probeSet().getActiveProbesForChromosome(viewer.chromosome());
//		System.out.println("Chr"+viewer.chromosome().name()+" has "+probes.length+" probes on it");
		addMouseMotionListener(new FeatureListener());
		addMouseListener(new FeatureListener());
	}
		
	public void paintComponent (Graphics g) {
		super.paintComponent(g);
		drawnProbes = new Vector();
		fullWidth = getWidth();
		height = getHeight();
		maxLogRatio = viewer.maxLogRatio();
		minLogRatio = viewer.minLogRatio();
		Rectangle r = getVisibleRect();
		x = r.x;
		width = r.width;
		
		if (viewer.application().dataCollection().getActiveDataStore() == data) {
			g.setColor(Color.YELLOW);			
		}
		else if (viewer.getIndex(this)%2 == 0) {
			g.setColor(new Color(255,180,180));
		}
		else {
			g.setColor(new Color(255,220,220));
		}
		
		g.fillRect(x,0,width,height);
		
		// Draw a mid-line for the ratio
		g.setColor(Color.LIGHT_GRAY);
		g.drawLine(x,height/2,x+width,height/2);
		
		if (viewer.makingSelection()) {
			int selStart = viewer.selectionStart();
			int selEnd = viewer.selectionEnd();
			int useStart = (selEnd > selStart) ? selStart : selEnd;
			int selWidth = selEnd-selStart;
			if (selWidth < 0) selWidth = 0-selWidth;
			g.setColor(Color.GREEN);
			g.fillRect(useStart,0,selWidth,height);
		}
				
		// Now go through all the features figuring out whether they
		// need to be displayed
		
		// We optimise by not drawing probes which overlap.  We need to keep
		// track of where the last probe ended.  We need to reset this value
		// before starting to do the drawing.
		lastXStart = 0;
		
		int startBp = pixelToBp(x);
		int endBp = pixelToBp(x+width);
		for (int i=0;i<probes.length;i++) {
//			System.out.println("Looking at feature "+featureName+" Start: "+features[i].location().start()+" End: "+features[i].location().end()+" Start:"+startBp+" End:"+endBp);
			if (probes[i].position() > startBp && probes[i].position() < endBp) {
				drawProbe(probes[i],g);
			}
		}
		g.setColor(Color.GRAY);
		g.drawString(data.name(),x+2,height-2);
	}
	
	private void drawProbe (Probe p, Graphics g) {
						
		int wholeXStart = bpToPixel(p.position()-10);
		int wholeXEnd = bpToPixel(p.position()+10);
		if ((wholeXEnd-wholeXStart)<2) {
			wholeXEnd = wholeXStart+2;
		}
		
		// Don't draw probes which overlap exactly with the last one
		if (wholeXStart < lastXStart) return;
		
		lastXStart = wholeXStart;
		
		drawnProbes.add(new DrawnProbe(wholeXStart, wholeXEnd, p));

//		double maxLogRatio = ChipMonkPreferences.getInstance().getMaxLogRatio();
//		double minLogRatio = ChipMonkPreferences.getInstance().getMinLogRatio();
		
		double lnRatio = Math.log(data.getRatioForProbe(p))/Math.log(2);

		// Restict the range we cover...
		if (lnRatio > maxLogRatio) lnRatio = maxLogRatio;
		if (lnRatio < minLogRatio) lnRatio = minLogRatio;
		
		if (p == activeProbe) {
			g.setColor(Color.WHITE);
		}
		else {
			g.setColor(ColourGradient.getColor(lnRatio,minLogRatio,maxLogRatio));
		}
		
		int yBoxStart = 0;
		int yBoxEnd = 0;
		
		if (lnRatio >=0) {
			yBoxStart = height/2 - ((int)(((double)height/2)*(lnRatio/maxLogRatio)));
			yBoxEnd = height/2;
		}
		else {
			yBoxStart = height/2;
			yBoxEnd = height/2 + ((int)(((double)height/2)*(lnRatio/minLogRatio))); 
		}
		
//		System.out.println("Drawing probe from x="+wholeXStart+" y="+yBoxStart+" width="+(wholeXEnd-wholeXStart)+" height="+(yBoxEnd-yBoxStart));
		g.fillRect(wholeXStart,yBoxStart,(wholeXEnd-wholeXStart),yBoxEnd-yBoxStart);

	}
	private int pixelToBp (int x) {
		return (int)(((double)x/fullWidth)*viewer.chromosome().length());
	}
	
	private int bpToPixel (int bp) {
		return (int)(((double)bp/viewer.chromosome().length())*fullWidth);
		
	}
	
	private class DrawnProbe {
		
		public int start;
		public int end;
		public Probe probe;
		
		public DrawnProbe (int start, int end, Probe probe) {
			this.start = start;
			this.end = end;
			this.probe = probe;
		}
		
		public boolean isInFeature (int x) {
			if (x >=start && x <= end) {
				return true;
			}
			else {
				return false;
			}
		}
	}
	
	private class FeatureListener implements MouseMotionListener, MouseListener {

		public void mouseDragged(MouseEvent me) {
			viewer.setSelectionEnd(me.getX());
		}

		public void mouseMoved(MouseEvent me) {
			int x = me.getX();
			Enumeration e = drawnProbes.elements();
			while (e.hasMoreElements()) {
				DrawnProbe p = (DrawnProbe)e.nextElement();
				if (p.isInFeature(x)) {
					if (activeProbe != p.probe) {
						viewer.application().setStatusText(" "+data.name()+" "+p.probe.name()+" "+p.probe.position()+"bp"+" Ratio="+data.getRatioForProbe(p.probe));
						activeProbe = p.probe;
						repaint();
					}
				}
			}
		}

		public void mouseClicked(MouseEvent me) {
			if ((me.getModifiers() & InputEvent.BUTTON3_MASK) == InputEvent.BUTTON3_MASK) {
				viewer.zoomOut();
				return;
			}
		}
		
		public void mousePressed(MouseEvent me) {
			viewer.setMakingSelection(true);
			viewer.setSelectionStart(me.getX());
			viewer.setSelectionEnd(me.getX());
		}

		public void mouseReleased(MouseEvent me) {
			viewer.setMakingSelection(false);

			int width = viewer.selectionEnd() - viewer.selectionStart();
			if (width < 0) {
				width = 0-width;
			}
			if (width < 5) return;
			
						
			viewer.setView(pixelToBp(viewer.selectionStart()),pixelToBp(viewer.selectionEnd()));
		}

		public void mouseEntered(MouseEvent arg0) {
			viewer.application().setStatusText(" "+data.name());
		}

		public void mouseExited(MouseEvent arg0) {
			activeProbe = null;
			repaint();
		}
		
	}

}
